#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2020 Valve Corporation
# Copyright (c) 2018-2020 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import re
import sys
from khronos_base_generator import write
from copy import deepcopy


class KhronosStructDecodersBodyGenerator():
    """Base class for generating struct docoder body code."""

    def generate_struct_decoder_content(self):
        """Performs C++ code generation for the feature."""
        api_data = self.get_api_data()

        write(
            'size_t Decode{func}Struct(const uint8_t* buffer, size_t buffer_size, {node}Node** {var});'
            .format(
                func=api_data.extended_struct_func_prefix,
                node=api_data.extended_struct_node_prefix,
                var=api_data.extended_struct_variable
            ),
            file=self.outFile
        )

        for struct in self.get_all_filtered_struct_names():
            if struct in self.all_struct_aliases or struct in self.all_union_aliases:
                continue

            body = '\n'
            body += 'size_t DecodeStruct(const uint8_t* buffer, size_t buffer_size, Decoded_{}* wrapper)\n'.format(
                struct
            )
            body += '{\n'
            body += '    assert((wrapper != nullptr) && (wrapper->decoded_value != nullptr));\n'
            body += '\n'
            body += '    size_t bytes_read = 0;\n'
            body += '    {}* value = wrapper->decoded_value;\n'.format(struct)
            body += '\n'
            body += self.make_decode_struct_body(
                struct, self.all_struct_members[struct]
            )
            body += '\n'
            body += '    return bytes_read;\n'
            body += '}'

            write(body, file=self.outFile)

    def make_decode_struct_body(self, name, values):
        """Generate C++ code for the decoder method body."""
        preamble = ''
        main_body = ''
        epilogue = ''

        for value in values:
            # If it is an extended struct name, it requires special treatment
            if self.is_extended_struct_definition(value):
                extended_struct_name = self.get_extended_struct_var_name()
                extended_struct_func_prefix = self.get_extended_struct_func_prefix(
                )
                main_body += '    bytes_read += Decode{}Struct((buffer + bytes_read), (buffer_size - bytes_read), &(wrapper->{}));\n'.format(
                    extended_struct_func_prefix, value.name
                )
                main_body += '    value->{var} = wrapper->{var} ? wrapper->{var}->GetPointer() : nullptr;\n'.format(
                    var=extended_struct_name
                )
            else:
                preamble, main_body, epilogue = self.make_decode_invocation(
                    name, value, preamble, main_body, epilogue
                )

        if len(preamble) > 0:
            preamble += '\n'
        body = preamble + main_body + epilogue
        return body

    def make_decode_invocation(
        self, name, value, preamble, main_body, epilogue
    ):
        """Generate the struct member decoder function call invocation."""
        buffer_args = '(buffer + bytes_read), (buffer_size - bytes_read)'

        is_struct = False
        is_string = False
        is_funcp = False
        is_handle_like = False
        is_enum = False

        type_name = self.make_invocation_type_name(value.base_type)

        if self.is_struct(type_name):
            is_struct = True
        elif type_name in ['String', 'WString']:
            is_string = True
        elif type_name == 'FunctionPtr':
            is_funcp = True
        elif self.is_handle_like(value.base_type):
            is_handle_like = True
        elif type_name == 'Enum':
            is_enum = True

        # is_pointer will be False for static arrays.
        if value.is_pointer or value.is_array:
            if type_name in self.EXTERNAL_OBJECT_TYPES and not value.is_array:
                # Pointer to an unknown object type, encoded as a 64-bit integer ID.
                main_body += '    bytes_read += ValueDecoder::DecodeAddress({}, &(wrapper->{}));\n'.format(
                    buffer_args, value.name
                )
                main_body += '    value->{} = nullptr;\n'.format(value.name)
            else:
                is_static_array = True if (
                    value.is_array and not value.is_dynamic
                ) else False
                access_op = '.'

                if value.base_type in self.children_structs.keys():
                    if value.pointer_count == 1:
                        main_body += f'    bytes_read += wrapper->{value.name}->DecodeBaseHeader((buffer + bytes_read), (buffer_size - bytes_read));\n'
                    else:
                        main_body += '\n'
                        main_body += '    // For base header arrays of pointers, we need to allocate an array to the generic base type pointer first\n'
                        main_body += '    // and then read the array attributes so we can jump right in to decoding the contents\n'
                        main_body += f'    wrapper->{value.name} = DecodeAllocator::Allocate<StructPointerDecoder<Decoded_{value.base_type}*>>();\n'
                        main_body += f'    bytes_read += wrapper->{value.name}->DecodeBaseHeader((buffer + bytes_read), (buffer_size - bytes_read));\n'
                        main_body += f'    value->{value.name} = wrapper->{value.name}->GetPointer();\n'
                else:
                    if is_struct:
                        main_body += '    wrapper->{} = DecodeAllocator::Allocate<{}>();\n'.format(
                            value.name, self.make_decoded_param_type(value)
                        )
                        access_op = '->'

                    if is_static_array:
                        # The pointer decoder will write directly to the struct member's memory.
                        main_body += '    wrapper->{name}{}SetExternalMemory(value->{name}, {arraylen});\n'.format(
                            access_op,
                            name=value.name,
                            arraylen=value.array_capacity
                        )

                    if is_struct or is_string or is_handle_like:
                        main_body += '    bytes_read += wrapper->{}{}Decode({});\n'.format(
                            value.name, access_op, buffer_args
                        )
                    elif self.has_basetype(value.base_type):
                        base_type = self.get_basetype(value.base_type)
                        main_body += '    bytes_read += wrapper->{}.Decode{}({});\n'.format(
                            value.name, self.encode_types[base_type],
                            buffer_args
                        )
                    else:
                        main_body += '    bytes_read += wrapper->{}.Decode{}({});\n'.format(
                            value.name, type_name, buffer_args
                        )

                    if not is_static_array:
                        if is_handle_like:
                            # Point the real struct's member pointer to the handle pointer decoder's handle memory.
                            main_body += '    value->{} = nullptr;\n'.format(
                                value.name
                            )
                        else:
                            # Point the real struct's member pointer to the pointer decoder's memory.
                            convert_const_cast_begin = ''
                            convert_const_cast_end = ''

                            if value.full_type.find('LPCWSTR *') != -1:
                                convert_const_cast_end = ')'
                                convert_const_cast_begin = 'const_cast<LPCWSTR*>('

                            elif value.full_type.find('LPCSTR *') != -1:
                                convert_const_cast_end = ')'
                                convert_const_cast_begin = 'const_cast<LPCSTR*>('

                            main_body += '    value->{name} = {}wrapper->{name}{}GetPointer(){};\n'.format(
                                convert_const_cast_begin,
                                access_op,
                                convert_const_cast_end,
                                name=value.name
                            )
        else:
            if is_struct:
                if value.base_type in self.children_structs.keys():
                    main_body += f'        switch ({value.name}->GetPointer()->type)\n'
                    main_body += '        {\n'
                    main_body += '            default:\n'
                    main_body += '                wrapper->{} = DecodeAllocator::Allocate<{}>();\n'.format(
                        value.name, self.make_decoded_param_type(value)
                    )
                    main_body += '                wrapper->{name}->decoded_value = &(value->{name});\n'.format(
                        name=value.name
                    )
                    main_body += '                bytes_read += DecodeStruct({}, wrapper->{});\n'.format(
                        buffer_args, value.name
                    )
                    main_body += '                break;\n'
                    for child in self.children_structs[value.base_type]:
                        switch_type = self.struct_type_names[child]

                        new_value = deepcopy(value)
                        new_value.base_type = child

                        decode_type = self.make_decoded_param_type(new_value)
                        var_name = value.name + '_' + child.lower()
                        preamble += f'    {decode_type}* {var_name};\n'

                        main_body += f'            case {switch_type}:\n'

                        main_body += '            {} = DecodeAllocator::Allocate<{}>();\n'.format(
                            var_name, decode_type
                        )
                        main_body += '            bytes_read += {}->Decode({});\n'.format(
                            var_name, buffer_args
                        )
                        main_body += '            wrapper->{} = std::move(reinterpret_cast<StructPointerDecoder<Decoded_{}>*>({}));\n'.format(
                            value.name, value.base_type, var_name
                        )
                        main_body += '            break;\n'
                    main_body += '        }\n'
                else:
                    main_body += '    wrapper->{} = DecodeAllocator::Allocate<{}>();\n'.format(
                        value.name, self.make_decoded_param_type(value)
                    )
                    main_body += '    wrapper->{name}->decoded_value = &(value->{name});\n'.format(
                        name=value.name
                    )
                    main_body += '    bytes_read += DecodeStruct({}, wrapper->{});\n'.format(
                        buffer_args, value.name
                    )
            elif is_funcp:
                main_body += '    bytes_read += ValueDecoder::DecodeAddress({}, &(wrapper->{}));\n'.format(
                    buffer_args, value.name
                )
                main_body += '    value->{} = nullptr;\n'.format(value.name)
            elif is_handle_like:
                main_body += '    bytes_read += ValueDecoder::DecodeHandleIdValue({}, &(wrapper->{}));\n'.format(
                    buffer_args, value.name
                )
                default_type = self.get_default_handle_atom_value(
                    value.base_type
                )
                main_body += '    value->{} = {};\n'.format(
                    value.name, default_type
                )
            elif self.is_generic_struct_handle_value(name, value.name):
                main_body += '    bytes_read += ValueDecoder::DecodeUInt64Value({}, &(wrapper->{}));\n'.format(
                    buffer_args, value.name
                )
                main_body += '    value->{} = 0;\n'.format(value.name)
            elif value.bitfield_width:
                # Bit fields need to be read into a tempoaray and then assigned to the struct member.
                temp_param_name = 'temp_{}'.format(value.name)
                main_body += '    {} {};\n'.format(
                    value.base_type, temp_param_name
                )
                main_body += '    bytes_read += ValueDecoder::Decode{}Value({}, &{});\n'.format(
                    type_name, buffer_args, temp_param_name
                )
                main_body += '    value->{} = {};\n'.format(
                    value.name, temp_param_name
                )
            elif is_enum:
                main_body += '    bytes_read += ValueDecoder::DecodeEnumValue({}, &(value->{}));\n'.format(
                    buffer_args, value.name
                )
            elif self.has_basetype(type_name):
                base_type = self.get_basetype(type_name)
                main_body += '    bytes_read += ValueDecoder::Decode{}Value({}, &(value->{}));\n'.format(
                    self.encode_types[base_type], buffer_args, value.name
                )
            elif 'Flags' in type_name:
                if 'Flags64' in type_name:
                    main_body += '    bytes_read += ValueDecoder::DecodeFlags64Value({}, &(value->{}));\n'.format(
                        buffer_args, value.name
                    )
                else:
                    main_body += '    bytes_read += ValueDecoder::DecodeFlagsValue({}, &(value->{}));\n'.format(
                        buffer_args, value.name
                    )
            else:
                main_body += '    bytes_read += ValueDecoder::Decode{}Value({}, &(value->{}));\n'.format(
                    type_name, buffer_args, value.name
                )

        return preamble, main_body, epilogue



#!/usr/bin/python3 -i
#
# Copyright 2013-2024, The Khronos Group Inc.
#
# SPDX-License-Identifier: Apache-2.0

from generator import OutputGenerator, write
from spec_tools.attributes import ExternSyncEntry
from spec_tools.validity import ValidityCollection, ValidityEntry
from spec_tools.util import getElemName
from pathlib import Path


class HostSynchronizationOutputGenerator(OutputGenerator):
    """HostSynchronizationOutputGenerator - subclass of OutputGenerator.
    Generates AsciiDoc includes of the externsync parameter table for the
    fundamentals chapter of the API specification. Similar to
    DocOutputGenerator.

    ---- methods ----
    HostSynchronizationOutputGenerator(errFile, warnFile, diagFile) - args as for
      OutputGenerator. Defines additional internal state.
    ---- methods overriding base class ----
    genCmd(cmdinfo)"""
    # Generate Host Synchronized Parameters in a table at the top of the spec

    threadsafety = {
        'parameters': ValidityCollection(),
        'parameterlists': ValidityCollection(),
        'implicit': ValidityCollection()
    }

    def makeParameterName(self, name):
        return f"pname:{name}"

    def makeFLink(self, name):
        return f"flink:{name}"

    def writeBlock(self, basename, title, contents):
        """Generate an include file.

        - directory - subdirectory to put file in
        - basename - base name of the file
        - contents - contents of the file (Asciidoc boilerplate aside)"""
        assert self.genOpts
        filename = Path(self.genOpts.directory) / basename
        self.logMsg('diag', '# Generating include file:', filename)
        with open(filename, 'w', encoding='utf-8') as fp:
            write(self.genOpts.conventions.warning_comment, file=fp)

            if contents:
                write(f'.{title}', file=fp)
                write('****', file=fp)
                write(contents, file=fp, end='')
                write('****', file=fp)
                write('', file=fp)
            else:
                self.logMsg('diag', '# No contents for:', filename)

    def writeInclude(self):
        "Generates the asciidoc include files."""
        assert self.genOpts
        file_suffix = self.genOpts.conventions.file_suffix
        self.writeBlock(f'parameters{file_suffix}',
                        'Externally Synchronized Parameters',
                        self.threadsafety['parameters'])
        self.writeBlock(f'parameterlists{file_suffix}',
                        'Externally Synchronized Parameter Lists',
                        self.threadsafety['parameterlists'])
        self.writeBlock(f'implicit{file_suffix}',
                        'Implicit Externally Synchronized Parameters',
                        self.threadsafety['implicit'])

    def makeThreadSafetyBlocks(self, cmd, paramtext):
        # See also makeThreadSafetyBlock in validitygenerator.py - similar but not entirely identical
        protoname = cmd.find('proto/name').text

        # Find and add any parameters that are thread unsafe
        explicitexternsyncparams = cmd.findall(f"{paramtext}[@externsync]")
        if explicitexternsyncparams is not None:
            for param in explicitexternsyncparams:
                self.makeThreadSafetyForParam(protoname, param)

        # Find and add any "implicit" parameters that are thread unsafe
        implicitexternsyncparams = cmd.find('implicitexternsyncparams')
        if implicitexternsyncparams is not None:
            for elem in implicitexternsyncparams:
                entry = ValidityEntry()
                entry += elem.text
                entry += ' in '
                entry += self.makeFLink(protoname)
                self.threadsafety['implicit'] += entry

        # Add a VU for any command requiring host synchronization.
        # This could be further parameterized, if a future non-Vulkan API
        # requires it.
        if self.genOpts.conventions.is_externsync_command(protoname):
            entry = ValidityEntry()
            entry += 'The sname:VkCommandPool that pname:commandBuffer was allocated from, in '
            entry += self.makeFLink(protoname)
            self.threadsafety['implicit'] += entry

    def makeThreadSafetyForParam(self, protoname, param):
        """Create thread safety validity for a single param of a command."""
        externsyncattribs = ExternSyncEntry.parse_externsync_from_param(param)
        param_name = getElemName(param)

        for attrib in externsyncattribs:
            entry = ValidityEntry()
            is_array = False
            if attrib.entirely_extern_sync:
                # "true" or "true_with_children"
                if self.paramIsArray(param):
                    entry += 'Each element of the '
                    is_array = True
                elif self.paramIsPointer(param):
                    entry += 'The object referenced by the '
                else:
                    entry += 'The '

                entry += self.makeParameterName(param_name)
                entry += ' parameter'

                if attrib.children_extern_sync:
                    entry += ', and any child handles,'

            else:
                # parameter/member reference
                readable = attrib.get_human_readable(make_param_name=self.makeParameterName)
                is_array = (' element of ' in readable)
                entry += readable

            entry += ' in '
            entry += self.makeFLink(protoname)

            if is_array:
                self.threadsafety['parameterlists'] += entry
            else:
                self.threadsafety['parameters'] += entry

    def genCmd(self, cmdinfo, name, alias):
        "Generate command."
        OutputGenerator.genCmd(self, cmdinfo, name, alias)

        # @@@ (Jon) something needs to be done here to handle aliases, probably

        self.makeThreadSafetyBlocks(cmdinfo.elem, 'param')

        self.writeInclude()



#!/usr/bin/env python3
# Copyright (c) 2019-2024, The Khronos Group Inc.
# SPDX-License-Identifier: Apache-2.0

from itertools import product
from pathlib import Path
import shutil
import sys

from shared import PLATFORMS, TRUE_FALSE, VS_VERSION, make_win_artifact_name

CWD = Path.cwd()


def move(src, dest):

    print(str(src), '->', str(dest))
    src.replace(dest)


if __name__ == "__main__":

    configs = {}
    workspace = Path(sys.argv[1])
    outbase = Path(sys.argv[2])

    common_copied = False

    for platform, uwp in product(PLATFORMS, TRUE_FALSE):
        # ARM/ARM64 is only built for the UWP platform.
        if not uwp and (platform.lower() == 'arm' or platform.lower() == 'arm64'):
            continue

        platform_dirname = '{}{}'.format(platform,
                                         '_uwp' if uwp else '')

        name = make_win_artifact_name(platform, uwp)

        artifact = workspace / name

        if not common_copied:
            # Start by copying the full tree over.
            shutil.copytree(artifact, outbase, dirs_exist_ok=True)
            common_copied = True
            continue

        # lib files
        shutil.copytree(artifact / platform_dirname, outbase / platform_dirname, dirs_exist_ok=True)



#!/usr/bin/env python
#
# Copyright 2009, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tests the --help flag of Google C++ Testing and Mocking Framework.

SYNOPSIS
       gtest_help_test.py --build_dir=BUILD/DIR
         # where BUILD/DIR contains the built gtest_help_test_ file.
       gtest_help_test.py
"""

import os
import re
import gtest_test_utils


IS_LINUX = os.name == 'posix' and os.uname()[0] == 'Linux'
IS_GNUKFREEBSD = os.name == 'posix' and os.uname()[0] == 'GNU/kFreeBSD'
IS_WINDOWS = os.name == 'nt'

PROGRAM_PATH = gtest_test_utils.GetTestExecutablePath('gtest_help_test_')
FLAG_PREFIX = '--gtest_'
DEATH_TEST_STYLE_FLAG = FLAG_PREFIX + 'death_test_style'
STREAM_RESULT_TO_FLAG = FLAG_PREFIX + 'stream_result_to'
UNKNOWN_FLAG = FLAG_PREFIX + 'unknown_flag_for_testing'
LIST_TESTS_FLAG = FLAG_PREFIX + 'list_tests'
INCORRECT_FLAG_VARIANTS = [re.sub('^--', '-', LIST_TESTS_FLAG),
                           re.sub('^--', '/', LIST_TESTS_FLAG),
                           re.sub('_', '-', LIST_TESTS_FLAG)]
INTERNAL_FLAG_FOR_TESTING = FLAG_PREFIX + 'internal_flag_for_testing'

SUPPORTS_DEATH_TESTS = "DeathTest" in gtest_test_utils.Subprocess(
    [PROGRAM_PATH, LIST_TESTS_FLAG]).output

# The help message must match this regex.
HELP_REGEX = re.compile(
    FLAG_PREFIX + r'list_tests.*' +
    FLAG_PREFIX + r'filter=.*' +
    FLAG_PREFIX + r'also_run_disabled_tests.*' +
    FLAG_PREFIX + r'repeat=.*' +
    FLAG_PREFIX + r'shuffle.*' +
    FLAG_PREFIX + r'random_seed=.*' +
    FLAG_PREFIX + r'color=.*' +
    FLAG_PREFIX + r'brief.*' +
    FLAG_PREFIX + r'print_time.*' +
    FLAG_PREFIX + r'output=.*' +
    FLAG_PREFIX + r'break_on_failure.*' +
    FLAG_PREFIX + r'throw_on_failure.*' +
    FLAG_PREFIX + r'catch_exceptions=0.*',
    re.DOTALL)


def RunWithFlag(flag):
  """Runs gtest_help_test_ with the given flag.

  Returns:
    the exit code and the text output as a tuple.
  Args:
    flag: the command-line flag to pass to gtest_help_test_, or None.
  """

  if flag is None:
    command = [PROGRAM_PATH]
  else:
    command = [PROGRAM_PATH, flag]
  child = gtest_test_utils.Subprocess(command)
  return child.exit_code, child.output


class GTestHelpTest(gtest_test_utils.TestCase):
  """Tests the --help flag and its equivalent forms."""

  def TestHelpFlag(self, flag):
    """Verifies correct behavior when help flag is specified.

    The right message must be printed and the tests must
    skipped when the given flag is specified.

    Args:
      flag:  A flag to pass to the binary or None.
    """

    exit_code, output = RunWithFlag(flag)
    self.assertEquals(0, exit_code)
    self.assert_(HELP_REGEX.search(output), output)

    if IS_LINUX or IS_GNUKFREEBSD:
      self.assert_(STREAM_RESULT_TO_FLAG in output, output)
    else:
      self.assert_(STREAM_RESULT_TO_FLAG not in output, output)

    if SUPPORTS_DEATH_TESTS and not IS_WINDOWS:
      self.assert_(DEATH_TEST_STYLE_FLAG in output, output)
    else:
      self.assert_(DEATH_TEST_STYLE_FLAG not in output, output)

  def TestNonHelpFlag(self, flag):
    """Verifies correct behavior when no help flag is specified.

    Verifies that when no help flag is specified, the tests are run
    and the help message is not printed.

    Args:
      flag:  A flag to pass to the binary or None.
    """

    exit_code, output = RunWithFlag(flag)
    self.assert_(exit_code != 0)
    self.assert_(not HELP_REGEX.search(output), output)

  def testPrintsHelpWithFullFlag(self):
    self.TestHelpFlag('--help')

  def testPrintsHelpWithShortFlag(self):
    self.TestHelpFlag('-h')

  def testPrintsHelpWithQuestionFlag(self):
    self.TestHelpFlag('-?')

  def testPrintsHelpWithWindowsStyleQuestionFlag(self):
    self.TestHelpFlag('/?')

  def testPrintsHelpWithUnrecognizedGoogleTestFlag(self):
    self.TestHelpFlag(UNKNOWN_FLAG)

  def testPrintsHelpWithIncorrectFlagStyle(self):
    for incorrect_flag in INCORRECT_FLAG_VARIANTS:
      self.TestHelpFlag(incorrect_flag)

  def testRunsTestsWithoutHelpFlag(self):
    """Verifies that when no help flag is specified, the tests are run
    and the help message is not printed."""

    self.TestNonHelpFlag(None)

  def testRunsTestsWithGtestInternalFlag(self):
    """Verifies that the tests are run and no help message is printed when
    a flag starting with Google Test prefix and 'internal_' is supplied."""

    self.TestNonHelpFlag(INTERNAL_FLAG_FOR_TESTING)


if __name__ == '__main__':
  gtest_test_utils.Main()



#!/usr/bin/python3 -i
#
# Copyright (c) 2022-2023 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, write
from khronos_json_consumer_body_generator import KhronosExportJsonConsumerBodyGenerator
from reformat_code import format_cpp_code, indent_cpp_code, remove_trailing_newlines


class VulkanExportJsonConsumerBodyGeneratorOptions(VulkanBaseGeneratorOptions, KhronosExportJsonConsumerBodyGenerator):
    """Options for generating a C++ class for Vulkan capture file to JSON file generation."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.extend((
            'util/defines.h',
            'generated/generated_vulkan_json_consumer.h',
            'decode/custom_vulkan_struct_to_json.h',
        ))
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'decode'))


class VulkanExportJsonConsumerBodyGenerator(VulkanBaseGenerator, KhronosExportJsonConsumerBodyGenerator):
    """VulkanExportJsonConsumerBodyGenerator - subclass of VulkanBaseGenerator.
    Generates C++ member definitions for the VulkanExportJsonConsumer class responsible for
    generating a textfile containing decoded Vulkan API call parameter data.
    Generate a C++ class for Vulkan capture file to JSON file generation.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        self.customImplementationRequired = {
            'vkCmdBuildAccelerationStructuresIndirectKHR',
            'vkCmdPushConstants',
            'vkCreatePipelineCache',
            'vkCreateShaderModule',
            'vkGetPipelineCacheData',
        }

        self.formatAsHex = {
            'VkDeviceAddress',
        }

        # Parameters using this name should be output as handles even though they are uint64_t
        self.formatAsHandle = {
            'objectHandle',
        }

        self.queueSubmit = {
            "vkQueueSubmit",
            "vkQueueSubmit2",
            "vkQueuePresentKHR",
            "vkQueueSubmit2KHR",
            }

    def endFile(self):
        """Method override."""
        # TODO: Each code generator is passed a blacklist like framework\generated\vulkan_generators\blacklists.json
        # of functions and structures not to generate code for. Once the feature is implemented, the following can be
        # replaced with adding vkCreateRayTracingPipelinesKHR in corresponding blacklist.
        if 'vkCreateRayTracingPipelinesKHR' in self.APICALL_BLACKLIST:
            self.APICALL_BLACKLIST.remove('vkCreateRayTracingPipelinesKHR')

        KhronosExportJsonConsumerBodyGenerator.generate_json_content(self)

        # Finish processing in superclass
        VulkanBaseGenerator.endFile(self)

    def need_feature_generation(self):
        """Indicates that the current feature has C++ code to generate."""
        if self.feature_cmd_params:
            return True
        return False

    def is_command_buffer_cmd(self, command):
        if 'vkCmd' in command:
            return True
        return False

    def skip_generating_command_json(self, command):
        """Method override"""
        return command in self.customImplementationRequired

    def decode_as_handle(self, value):
        """Method override
        Indicates that the given type should be decoded as a handle."""
        return (
            (
                self.is_handle_like(value.base_type)
                or value.name in self.formatAsHandle
            )
        )

    def decode_as_hex(self, value):
        """Method override"""
        return value.base_type in self.formatAsHex

    # yapf: disable
    def make_consumer_func_body(self, return_type, name, values):
        """Return class member function definition."""
        body = ''

        if name in self.queueSubmit:
            body += '    FieldToJson(jdata[NameSubmitIndex()], ++submit_index_, json_options);\n'
        elif self.is_command_buffer_cmd(name):
            body += '    FieldToJson(jdata[NameCommandIndex()], GetCommandBufferRecordIndex(commandBuffer), json_options);\n'

        body += KhronosExportJsonConsumerBodyGenerator.make_consumer_func_body(self, return_type, name, values)
        return body
    # yapf: enable



#!/usr/bin/env python3
#
# Copyright (c) 2021-2023 LunarG, Inc.
# Copyright (c) 2023 Valve Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from dx12_base_generator import Dx12BaseGenerator, write
from dx12_enum_to_string_header_generator import Dx12EnumToStringHeaderGenerator
from reformat_code import format_cpp_code

class Dx12EnumToJsonHeaderGenerator(Dx12BaseGenerator):
    """Generates C++ functions which build JSON representations of enums."""

    BITS_LIST = Dx12EnumToStringHeaderGenerator.BITS_LIST

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )

    def beginFile(self, gen_opts):
        """Method override."""
        Dx12BaseGenerator.beginFile(self, gen_opts)

        code = format_cpp_code('''
            #if defined(D3D12_SUPPORT)

            /// @file Functions to convert enums to JSON. While trivial these do tidy up
            /// the FieldToJsons of structs which use them and the JSON consumer too.
            /// They also mean that generators don't need separate cases for enums.

            #include "util/json_util.h"
            #include "generated/generated_dx12_enum_to_string.h"
            #include "util/defines.h"
            #include "util/to_string.h"
            #include "format/platform_types.h"

            GFXRECON_BEGIN_NAMESPACE(gfxrecon)
            GFXRECON_BEGIN_NAMESPACE(util)
        ''')
        write(code, file=self.outFile)
        self.newline()

    def generate_feature(self):
        enum_dict = self.source_dict['enum_dict']
        enum_prototypes = ''
        flag_prototypes = ''

        for k, v in enum_dict.items():
            # Generate enum handler for all enums
            enum_prototypes += format_cpp_code('''inline void FieldToJson(nlohmann::ordered_json& jdata, const {0} value, const JsonOptions& options = JsonOptions())
            {{
                FieldToJson(jdata, ToString(value), options);
            }}
            inline void FieldToJson(nlohmann::ordered_json& jdata, const {0}* pEnum, const JsonOptions& options = JsonOptions())
            {{
                FieldToJson(jdata, *pEnum, options);
            }}
            '''.format(k))
            enum_prototypes += '\n\n'

            # Generate flags handler for enums identified as bitmasks
            for bits in self.BITS_LIST:
                if k.find(bits) >= 0:
                    flag_prototypes += format_cpp_code('''inline void FieldToJson_{0}(nlohmann::ordered_json& jdata, const uint32_t flags, const JsonOptions& options = JsonOptions())
                    {{
                        std::string representation;
                        if (!options.expand_flags)
                        {{
                            representation = to_hex_fixed_width(flags);
                        }}
                        else
                        {{
                            representation = ToString_{0}(flags);
                        }}
                        FieldToJson(jdata, representation, options);
                    }}
                    \n'''.format(k))
                    flag_prototypes += '\n'

        write(enum_prototypes, file=self.outFile)
        write(flag_prototypes, file=self.outFile)

        write(format_cpp_code('''
        // IID struct-as-enum special case:
        inline void FieldToJson(nlohmann::ordered_json& jdata, const IID& value, const JsonOptions& options = JsonOptions())
        {
            FieldToJson(jdata, ToString(value), options);
        }
        '''), file=self.outFile)

    def endFile(self):
        """Method override."""
        self.newline()
        write('GFXRECON_END_NAMESPACE(util)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)
        self.newline()
        write('#endif // defined(D3D12_SUPPORT)', file=self.outFile)

        # Finish processing in superclass
        Dx12BaseGenerator.endFile(self)



#!/usr/bin/python3 -i
#
# Copyright (c) 2021-2022 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
import inspect
from openxr_base_generator import *
from khronos_enum_to_string_body_generator import KhronosEnumToStringBodyGenerator


class OpenXrEnumToStringBodyGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Options for generating C++ functions for OpenXr ToString() functions"""

    def __init__(
        self,
        # Path to JSON file listing apicalls and structs to ignore.
        blacklists=None,
        # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        platform_types=None,
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.append('generated_openxr_enum_to_string.h')
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'util'))
        self.begin_end_file_data.common_api_headers = []

# OpenXrEnumToStringBodyGenerator - subclass of OpenXrBaseGenerator.
# Generates C++ functions for stringifying OpenXr API enums.
class OpenXrEnumToStringBodyGenerator(OpenXrBaseGenerator, KhronosEnumToStringBodyGenerator):
    """Generate C++ functions for OpenXr ToString() functions"""

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    # Method override
    def endFile(self):
        KhronosEnumToStringBodyGenerator.write_enum_to_string_body(self, use_flags_for_64bit_enum=True)

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)

    def skip_generating_enum_to_string_for_type(self, enum_name):
        """ Override for OpenXr to suppress Vulkan types """
        return enum_name.startswith('Vk') or ('Bits' in enum_name and not self.is_flags_enum_64bit(enum_name))


#!/usr/bin/python3 -i
#
# Copyright (c) 2018 Valve Corporation
# Copyright (c) 2018 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from openxr_base_generator import OpenXrBaseGenerator, OpenXrBaseGeneratorOptions, write
from khronos_decode_extended_struct_generator import KhronosDecodeExtendedStructGenerator


class OpenXrStructNextDecodersGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Eliminates JSON black_lists and platform_types files, which are not necessary for
    pNext switch statement generation.
    Options for OpenXr API pNext structure decoding C++ code generation.
    """

    def __init__(
        self,
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            None,
            None,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )


class OpenXrStructNextDecodersGenerator(OpenXrBaseGenerator, KhronosDecodeExtendedStructGenerator):
    """DecodePNextStructGenerator - subclass of OpenXrBaseGenerator.
    Generates C++ code for OpenXr API pNext structure decoding.
    Generate pNext structure decoding C++ code.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def beginFile(self, gen_opts):
        """Method override."""
        OpenXrBaseGenerator.beginFile(self, gen_opts)

        KhronosDecodeExtendedStructGenerator.write_common_headers(self, gen_opts)

        self.newline()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(decode)', file=self.outFile)
        self.newline()

        KhronosDecodeExtendedStructGenerator.write_decode_struct_definition_prefix(self)

    def endFile(self):
        """Method override."""
        """Performs C++ code generation for the feature."""
        KhronosDecodeExtendedStructGenerator.write_decode_struct_definition_data(self)
        self.newline()
        write('GFXRECON_END_NAMESPACE(decode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)




#!/usr/bin/env python3
#
# Copyright (c) 2021 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from dx12_base_generator import Dx12BaseGenerator, write


# Generates declarations for functions to add entries for DX12 state tracker
# based on IID.
class Dx12AddEntriesHeaderGenerator(Dx12BaseGenerator):
    # Default C++ code indentation size.
    INDENT_SIZE = 4

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )

        # Unique set of names of all defined classes.
        self.class_names = []
        # Unique set of names of all class names specified as base classes.
        self.class_parent_names = []

    # Method override
    def beginFile(self, genOpts):
        Dx12BaseGenerator.beginFile(self, genOpts)
        self.write_include()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(encode)', file=self.outFile)

    # Method override
    def endFile(self):
        self.generate_all()
        indent = ''
        final_class_names = self.get_final_class_names()

        write('template <typename ParentWrapper>', file=self.outFile)
        write(
            'const std::unordered_map<IID, std::function<void(typename void**, format::ApiCallId, ParentWrapper*, const util::MemoryOutputStream*, std::mutex &state_table_mutex, Dx12StateTable &state_table)>,IidHash> kAddEntryFunctionTable',
            file=self.outFile
        )
        write('{', file=self.outFile)
        indent = self.increment_indent(indent)
        for final_class_name in final_class_names:
            class_family_names = self.get_class_family_names(final_class_name)
            for name in class_family_names:
                code = indent
                code += '{ IID_'
                code += name + ', ' + self.gen_create_func_name(
                    class_family_names
                )
                code += ' },'
                write(code, file=self.outFile)
        indent = self.decrement_indent(indent)
        write('};', file=self.outFile)
        self.newline()

        write(
            'const std::unordered_map<IID, std::function<void(typename void**, format::ApiCallId, void*, const util::MemoryOutputStream*, std::mutex &state_table_mutex, Dx12StateTable &state_table)>,IidHash> kAddEntryVoidFunctionTable',
            file=self.outFile
        )
        write('{', file=self.outFile)
        indent = self.increment_indent(indent)
        for final_class_name in final_class_names:
            class_family_names = self.get_class_family_names(final_class_name)
            for name in class_family_names:
                code = indent
                code += '{ IID_'
                code += name + ', ' + self.gen_create_void_func_name(
                    class_family_names
                )
                code += ' },'
                write(code, file=self.outFile)
        indent = self.decrement_indent(indent)
        write('};', file=self.outFile)
        self.newline()

        write(
            'static DxWrapperInfo* GetWrapperInfo(IUnknown_Wrapper* wrapper)',
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    GFXRECON_ASSERT(wrapper != nullptr);', file=self.outFile)
        write('    auto riid = wrapper->GetRiid();', file=self.outFile)
        for final_class_name in final_class_names:
            class_family_names = self.get_class_family_names(final_class_name)
            for name in class_family_names:
                code = '    if(riid == IID_' + name + ')\n'
                code += '    {\n'
                code += '        auto* new_wrapper = reinterpret_cast<' + class_family_names[
                    0] + '_Wrapper*>(wrapper);\n'
                code += '        return new_wrapper->GetObjectInfo().get();\n'
                code += '    }'
                write(code, file=self.outFile)
        write('    return nullptr;', file=self.outFile)
        write('}', file=self.outFile)
        self.newline()

        write('GFXRECON_END_NAMESPACE(encode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        Dx12BaseGenerator.endFile(self)

    # Method override
    def generate_feature(self):
        Dx12BaseGenerator.generate_feature(self)

        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            for class_name, class_value in v.classes.items():
                if self.is_required_class_data(class_value)\
                   and (class_value['name'] != 'IUnknown'):
                    # Track class names
                    if class_name not in self.class_names:
                        self.class_names.append(class_name)

                    # Track names of classes inherited from
                    for entry in class_value['inherits']:
                        decl_name = entry['decl_name']
                        if decl_name not in self.class_parent_names:
                            self.class_parent_names.append(decl_name)

    # Get the names of the final classes in the DX class hierarchies.
    def get_final_class_names(self):
        final_class_names = []

        for name in self.class_names:
            if name not in self.class_parent_names:
                final_class_names.append(name)

        return final_class_names

    def get_class_family_names(self, final_class_name):
        base_name = final_class_name
        final_number = ''

        # Get the number from the end of the class name.  Start from the
        # back of the string and advance forward until a non-digit character
        # is encountered.
        if (not final_class_name
            in self.NOT_FAMILY_CLASSES) and (final_class_name[-1].isdigit()):
            for i in range(len(final_class_name) - 1, -1, -1):
                if not final_class_name[i].isdigit():
                    base_name = final_class_name[:i + 1]
                    final_number = final_class_name[i + 1:]
                    break

        class_family_names = []
        if base_name in self.FAMILY_CLASSES_EXECPTION:
            class_family_names.append(self.FAMILY_CLASSES_EXECPTION[base_name])
        class_family_names.append(base_name)
        if final_number:
            # Generate with class numbers in ascending order, from 1 to n.
            for i in range(1, int(final_number) + 1):
                class_family_names.append(base_name + str(i))

        return class_family_names

    def gen_create_func_name(self, class_family_names):
        # Create a function name and argument with the first class
        # name from the class family (the first name in the list is
        # the name without a version number).
        first_class = class_family_names[0]
        return 'AddEntry<{0}_Wrapper, ParentWrapper>'.format(first_class)

    def gen_create_void_func_name(self, class_family_names):
        # Create a function name and argument with the first class
        # name from the class family (the first name in the list is
        # the name without a version number).
        first_class = class_family_names[0]
        return 'AddEntry<{0}_Wrapper>'.format(first_class)

    def generate_all(self):
        self.newline()

        decl = 'template <typename Wrapper, typename ParentWrapper>\n'
        decl += 'void AddEntry(typename void** new_handle, format::ApiCallId create_call_id, ParentWrapper* create_object_wrapper, const util::MemoryOutputStream* create_parameter_buffer, std::mutex &state_table_mutex, Dx12StateTable &state_table)\n'
        decl += '{\n'
        decl += '    assert(create_object_wrapper != nullptr);\n'
        decl += '    if (*new_handle != nullptr)\n'
        decl += '    {\n'
        decl += '        auto wrapper = reinterpret_cast<Wrapper*>(*new_handle);\n'
        decl += '\n'
        decl += '        // Adds the handle wrapper to the object state table, filtering for duplicate handle retrieval.\n'
        decl += '        std::unique_lock<std::mutex> lock(state_table_mutex);\n'
        decl += '        if (state_table.InsertWrapper(wrapper->GetCaptureId(), wrapper))\n'
        decl += '        {\n'
        decl += '            dx12_state_tracker::InitializeState<Wrapper, ParentWrapper>(wrapper, create_call_id, create_object_wrapper, create_parameter_buffer);\n'
        decl += '        }\n'
        decl += '    }\n'
        decl += '}\n'
        decl += '\n'
        decl += 'template <typename Wrapper>\n'
        decl += 'void AddEntry(typename void** new_handle, format::ApiCallId create_call_id, void* create_object_wrapper, const util::MemoryOutputStream* create_parameter_buffer, std::mutex &state_table_mutex, Dx12StateTable &state_table)\n'
        decl += '{\n'
        decl += '    assert(create_object_wrapper == nullptr);\n'
        decl += '    if (*new_handle != nullptr)\n'
        decl += '    {\n'
        decl += '        auto wrapper = reinterpret_cast<Wrapper*>(*new_handle);\n'
        decl += '\n'
        decl += '        // Adds the handle wrapper to the object state table, filtering for duplicate handle retrieval.\n'
        decl += '        std::unique_lock<std::mutex> lock(state_table_mutex);\n'
        decl += '        if (state_table.InsertWrapper(wrapper->GetCaptureId(), wrapper))\n'
        decl += '        {\n'
        decl += '            dx12_state_tracker::InitializeState<Wrapper>(wrapper, create_call_id, static_cast<void*>(nullptr), create_parameter_buffer);\n'
        decl += '        }\n'
        decl += '    }\n'
        decl += '}\n'
        decl += '\n'
        write(decl, file=self.outFile)

    def write_include(self):
        write(
            '#include "generated/generated_dx12_wrapper_creators.h"\n',
            file=self.outFile
        )

    def increment_indent(self, indent):
        return indent + (' ' * self.INDENT_SIZE)

    def decrement_indent(self, indent):
        return indent[:-self.INDENT_SIZE]



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2020 Valve Corporation
# Copyright (c) 2018-2025 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from openxr_base_generator import OpenXrBaseGenerator, OpenXrBaseGeneratorOptions, write
from khronos_decoder_body_generator import KhronosDecoderBodyGenerator


class OpenXrDecoderBodyGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Options for generating a C++ class for OpenXr API parameter decoding."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[],
        extra_manual_commands=[],
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers,
            extra_manual_commands=extra_manual_commands,
        )

        self.begin_end_file_data.specific_headers.extend((
            'decode/handle_pointer_decoder.h',
            'decode/pointer_decoder.h',
            'decode/string_array_decoder.h',
            'decode/string_decoder.h',
            'decode/struct_pointer_decoder.h',
            'decode/value_decoder.h',
            'decode/openxr_next_node.h',
            'generated/generated_openxr_decoder.h',
            'generated/generated_openxr_struct_decoders_forward.h',
            'util/defines.h',
        ))
        self.begin_end_file_data.system_headers.append('cstddef')
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'decode'))

class OpenXrDecoderBodyGenerator(OpenXrBaseGenerator, KhronosDecoderBodyGenerator):
    """OpenXrDecoderBodyGenerator - subclass of OpenXrBaseGenerator.
    Generates C++ member functions for the OpenXrDecoder class responsible for decoding
    OpenXr API call parameter data.
    Generate a C++ class for OpenXr API parameter decoding.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        KhronosDecoderBodyGenerator.__init__(self)

        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def endFile(self):
        """Method override."""
        self.newline()
        self.generate_commands()

        self.newline()
        # Generate the OpenXrDecoder::DecodeFunctionCall method for all of the commands processed by the generator.
        self.generate_decode_cases()
        self.newline()

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)



#!/usr/bin/python3 -i
#
# Copyright (c) 2019 Valve Corporation
# Copyright (c) 2019-2025 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from khronos_base_generator import write


class KhronosStructHandleWrappersHeaderGenerator():
    """KhronosStructHandleWrappersHeaderGeneratorVulk
    Generates C++ function prototypes for wrapping struct member handles
    when recording Khronos API call parameter data.
    """

    def write_struct_handle_wrapper_content(self):
        # Generate unwrap and rewrap code for input structures.
        for struct in self.get_all_filtered_struct_names():
            if (
                (struct in self.structs_with_handles) or
                self.child_struct_has_handles(struct) or
                (struct in self.GENERIC_HANDLE_STRUCTS)
            ) and (struct not in self.STRUCT_MAPPERS_BLACKLIST):
                body = '\n'
                body += 'void UnwrapStructHandles({}* value, HandleUnwrapMemory* unwrap_memory);'.format(
                    struct
                )
                write(body, file=self.outFile)

        api_data = self.get_api_data()
        const_prefix = ''
        if api_data.return_const_ptr_on_extended:
            const_prefix = 'const '
        func_prefix = api_data.extended_struct_func_prefix
        in_struct   = api_data.base_in_struct

        lines = ['']
        lines.append(f'{in_struct}* Copy{func_prefix}Struct(const {in_struct}* base, HandleUnwrapMemory* unwrap_memory);')
        lines.append('')
        lines.append(f'{const_prefix}void* Unwrap{func_prefix}StructHandles(const void* value, HandleUnwrapMemory* unwrap_memory);')
        lines.append('')
        lines.append('template <typename ParentWrapper, typename CoParentWrapper, typename T>')
        lines.append( 'void CreateWrappedStructArrayHandles(typename ParentWrapper::HandleType parent, typename CoParentWrapper::HandleType co_parent, T* value, size_t len, PFN_GetHandleId get_id);')
        lines.append('')
        lines.append('template <typename T>')
        lines.append('T* MakeUnwrapStructs(const T* values, size_t len, HandleUnwrapMemory* unwrap_memory)')
        lines.append('{')
        lines.append('    assert((values != nullptr) && (len > 0) && (unwrap_memory != nullptr));')
        lines.append('')
        lines.append('    const uint8_t* bytes     = reinterpret_cast<const uint8_t*>(values);')
        lines.append('    size_t         num_bytes = len * sizeof(T);')
        lines.append('')
        lines.append('    return reinterpret_cast<T*>(unwrap_memory->GetFilledBuffer(bytes, num_bytes));')
        lines.append('}')
        lines.append('')
        lines.append('template <typename T>')
        lines.append(f'{const_prefix}T* UnwrapStructPtrHandles(const T* value, HandleUnwrapMemory* unwrap_memory)')
        lines.append('{')
        lines.append('    T* unwrapped_struct = nullptr;')
        lines.append('')
        lines.append('    if (value != nullptr)')
        lines.append('    {')
        lines.append('        unwrapped_struct = MakeUnwrapStructs(value, 1, unwrap_memory);')
        lines.append('        UnwrapStructHandles(unwrapped_struct, unwrap_memory);')
        lines.append('    }')
        lines.append('')
        lines.append('    return unwrapped_struct;')
        lines.append('}')
        lines.append('')
        write('\n'.join(lines), file=self.outFile)

        self.generate_create_wrapper_funcs()

        write(
            'template <typename ParentWrapper, typename CoParentWrapper, typename T>',
            file=self.outFile
        )
        write(
            'void CreateWrappedStructArrayHandles(typename ParentWrapper::HandleType parent, typename CoParentWrapper::HandleType co_parent, T* value, size_t len, PFN_GetHandleId get_id)',
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    if (value != nullptr)', file=self.outFile)
        write('    {', file=self.outFile)
        write('        for (size_t i = 0; i < len; ++i)', file=self.outFile)
        write('        {', file=self.outFile)
        write(
            '            CreateWrappedStructHandles<ParentWrapper, CoParentWrapper>(parent, co_parent, &value[i], get_id);',
            file=self.outFile
        )
        write('        }', file=self.outFile)
        write('    }', file=self.outFile)
        write('}', file=self.outFile)
        self.newline()

        write('template <typename T>', file=self.outFile)
        write(
            f'{const_prefix}T* UnwrapStructArrayHandles({const_prefix}T* values, size_t len, HandleUnwrapMemory* unwrap_memory)',
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    if ((values != nullptr) && (len > 0))', file=self.outFile)
        write('    {', file=self.outFile)
        write(
            '        auto unwrapped_structs = MakeUnwrapStructs(values, len, unwrap_memory);',
            file=self.outFile
        )
        self.newline()
        write('        for (size_t i = 0; i < len; ++i)', file=self.outFile)
        write('        {', file=self.outFile)
        write(
            '            UnwrapStructHandles(&unwrapped_structs[i], unwrap_memory);',
            file=self.outFile
        )
        write('        }', file=self.outFile)
        self.newline()
        write('        return unwrapped_structs;', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write(
            '    // Leave the original memory in place when the pointer is not null, but size is zero.',
            file=self.outFile
        )
        write('    return values;', file=self.outFile)
        write('}', file=self.outFile)
        self.newline()
        write('template <typename T>', file=self.outFile)
        write(
            'T* UnwrapStructPtrArrayHandles(T* values, size_t len, HandleUnwrapMemory* unwrap_memory)',
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    if ((values != nullptr) && (len > 0))', file=self.outFile)
        write('    {', file=self.outFile)
        write(
            '        auto unwrapped_structs = MakeUnwrapStructs(values, len, unwrap_memory);',
            file=self.outFile
        )
        self.newline()
        write('        for (size_t i = 0; i < len; ++i)', file=self.outFile)
        write('        {', file=self.outFile)
        write(
            '            UnwrapStructHandles(unwrapped_structs[i], unwrap_memory);',
            file=self.outFile
        )
        write('        }', file=self.outFile)
        self.newline()
        write('        return unwrapped_structs;', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write(
            '    // Leave the original memory in place when the pointer is not null, but size is zero.',
            file=self.outFile
        )
        write('    return values;', file=self.outFile)
        write('}', file=self.outFile)

    # Recursively search a structs members to see if they too belong in the
    # output struct list.  This could be because an including struct is an
    # output struct.
    def process_struct_members_to_output_struct(self, value):
        for member in self.all_struct_members[value.base_type]:
            if (
                self.is_struct(member.base_type)
                and not self.is_struct_black_listed(member.base_type)
                and (member.base_type in self.structs_with_handles)
                and (member.base_type not in self.output_structs)
            ):
                self.output_structs.append(member.base_type)
                self.process_struct_members_to_output_struct(member)

    def generate_create_wrapper_funcs(self):
        # Map of Vulkan structs containing handles to a list values for handle members or struct members
        # that contain handles (eg. VkGraphicsPipelineCreateInfo contains a VkPipelineShaderStageCreateInfo
        # member that contains handles).
        self.output_structs = [
        ]  # Output structures that retrieve handles, which need to be wrapped.

        # Check for output structures, which retrieve handles that need to be wrapped.
        for cmd in self.all_cmd_params:
            info = self.all_cmd_params[cmd]
            values = info[2]

            for value in values:
                if self.is_output_parameter(value) and self.is_struct(
                    value.base_type
                ) and (value.base_type in self.structs_with_handles
                       ) and (value.base_type not in self.output_structs):
                    self.output_structs.append(value.base_type)
                    self.process_struct_members_to_output_struct(value)

        """Generates functions that wrap struct handle members."""
        for struct in self.output_structs:
            body = 'template <typename ParentWrapper, typename CoParentWrapper>\n'
            body += 'void CreateWrappedStructHandles(typename ParentWrapper::HandleType parent, typename CoParentWrapper::HandleType co_parent, {}* value, PFN_GetHandleId get_id)\n'.format(
                struct
            )
            body += '{\n'
            body += '    if (value != nullptr)\n'
            body += '    {\n'

            default_info = {
                    'indent': '        ',
                    'struct': '',
                    'plural': '',
                    'length': '',
                    'coparent_wrapper': '',
                    'coparent_arg': '',
                    'dref_value': '&',
                }
            for member in self.structs_with_handles[struct]:
                # Set up the information needed to generation the createwrapper call
                info = { **default_info }
                info['prefix'] = self.get_wrapper_prefix_from_type(member.base_type)
                info['target'] = self.get_handle_like_func_name_modifier(member.base_type)
                info['name'] = member.name
                if not (self.is_atom(member.base_type) or self.is_opaque(member.base_type)):
                    info['coparent_wrapper'] = ', CoParentWrapper'
                    info['coparent_arg'] = 'co_parent, '

                if member.is_array or member.is_pointer:
                    info['dref_value'] = ''

                if self.is_struct(member.base_type):
                    info['type'] = member.base_type
                    info['plural'] = 's'
                    if member.is_array:
                        info['wrapper'] = ', ' + member.base_type
                        info['struct'] = 'StructArray'
                    else:
                        info['wrapper'] = ''
                        info['struct'] =  'Struct'
                else:
                    info['type'] = member.base_type[2:]
                    info['wrapper'] = ', {}::{}Wrapper'.format(info['prefix'], info['type'])

                if member.is_array:
                    info['length'] = 'value->{}, '.format(member.array_length)
                    info['plural'] = 's'

                wrapper_name ='        {prefix}::CreateWrapped{struct}{target}{plural}'.format(**info)
                wrapper_template = '<ParentWrapper{coparent_wrapper}{wrapper}>'.format(**info)
                wrapper_args ='(parent, {coparent_arg}{dref_value}value->{name}, {length}get_id);\n'.format(**info)
                body += wrapper_name + wrapper_template + wrapper_args

            body += '    }\n'
            body += '}\n'
            write(body, file=self.outFile)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018 Valve Corporation
# Copyright (c) 2018-2025 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from openxr_base_generator import OpenXrBaseGenerator, OpenXrBaseGeneratorOptions
from khronos_layer_func_table_generator import KhronosLayerFuncTableGenerator

class OpenXrLayerFuncTableGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Eliminates JSON black_lists and platform_types files, which are not necessary for
    function table generation .
    Options for OpenXR layer function table C++ code generation.
    """

    def __init__(
        self,
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            None,
            None,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )
        self.begin_end_file_data.specific_headers.extend((
            'encode/custom_openxr_api_call_encoders.h',
            'generated/generated_openxr_api_call_encoders.h',
            'layer/trace_layer.h',
            'util/defines.h',
        ))
        self.begin_end_file_data.system_headers.append('unordered_map')
        self.begin_end_file_data.namespaces.append('gfxrecon')

class OpenXrLayerFuncTableGenerator(OpenXrBaseGenerator, KhronosLayerFuncTableGenerator):
    """LayerFuncTableGenerator - subclass of BaseGenerator.
    Generates C++ function table for the OpenXR API calls exported by the layer.
    Generate OpenXr layer function table C++ type declarations.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        # API Calls that the trace layer should ignore
        self.APICALL_BLACKLIST += [
            'xrNegotiateLoaderRuntimeInterface',
            'xrNegotiateLoaderApiLayerInterface',
            'xrInitializeLoaderKHR',
            'xrCreateInstance',
            'xrCreateApiLayerInstance',
        ]

        # These functions are provided directly by the layer, and are not encoded
        self.LAYER_FUNCTIONS = [
            'xrEnumerateInstanceExtensionProperties',
            'xrEnumerateApiLayerProperties',
            'xrCreateApiLayerInstance',
            'xrGetInstanceProcAddr',
        ]

    def endFile(self):
        """Method override."""

        KhronosLayerFuncTableGenerator.write_layer_func_table_contents(self, self.LAYER_FUNCTIONS, 100)
        self.newline()

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)



#!/usr/bin/python3 -i
#
# Copyright (c) 2021 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from dx12_base_generator import Dx12BaseGenerator, write


class Dx12CommandListUtilHeaderGenerator(Dx12BaseGenerator):
    """Generate a C++ class for Dx12 Track CommandList objects.
    """

    # Tracked data is union. That's written in custom functions.
    SKIP_COMMAND_METHODS = [
        'ID3D12GraphicsCommandList_ResourceBarrier',
        'ID3D12GraphicsCommandList4_BeginRenderPass',
        'ID3D12GraphicsCommandList4_BuildRaytracingAccelerationStructure'
    ]

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )

    # Method override
    def beginFile(self, genOpts):
        Dx12BaseGenerator.beginFile(self, genOpts)
        self.write_include()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(encode)', file=self.outFile)

    def endFile(self):
        """Method override."""
        self.newline()
        write('GFXRECON_END_NAMESPACE(encode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        Dx12BaseGenerator.endFile(self)

    # Method override
    def generate_feature(self):
        Dx12BaseGenerator.generate_feature(self)
        self.newline()

        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            for class_name, class_value in v.classes.items():
                if self.is_required_class_data(
                    class_value
                ) and 'CommandList' in class_name:
                    for m in class_value['methods']['public']:
                        if not self.is_method_black_listed(
                            class_name, m['name']
                        ):
                            self.write_track_function(class_name, m)

    def write_track_function(self, class_name, method_info):
        if (
            class_name + '_' + method_info['name']
        ) in self.SKIP_COMMAND_METHODS:
            return

        parameters = 'ID3D12CommandList_Wrapper* wrapper'
        required = False

        for p in method_info['parameters']:
            value = self.get_value_info(p)
            value_type = self.clean_type_define(p['type'])

            if value.array_length and (type(value.array_length) == int):
                value_type += '*'

            parameters += ', ' + value_type + ' ' + value.name
            if self.is_track_command_list(value):
                required = True

        if required:
            code = 'void Track_{}_{}({})'.format(
                class_name, method_info['name'], parameters
            )

            code += self.get_track_function_body(class_name, method_info)

            write(code, file=self.outFile)
            self.newline()

    def get_track_function_body(self, class_name, method_info):
        return ';'

    def write_include(self):
        code = ''
        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            code += '#include <{}>\n'.format(k)

        code += (
            "\n"
            "#include \"generated/generated_dx12_wrappers.h\"\n"
            "\n"
            "#include \"util/defines.h\"\n"
            "\n"
        )
        write(code, file=self.outFile)



#!/usr/bin/python3
#
# Copyright (c) 2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from khronos_base_generator import write


class KhronosStructTypeUtilGenerator():
    """KhronosStructTypeUtilGenerator
    Generates C++ utility header to do compile-time lookups between Khronos API
    structure types and their corresponding enum values
    """

    @staticmethod
    def update_begin_end_file_data(begin_end):
        begin_end.specific_headers.extend((
        'format/platform_types.h',
        'util/defines.h',
        ))

        begin_end.namespaces.extend((
        'gfxrecon', 'util'
        ))

    def write_struct_type_prefix(self, gen_opts):
        self.newline()

        current_api_data = self.get_api_data()
        write(
            '// Instantiating the primary template indicates that either the template was',
            file=self.outFile
        )
        write(
            '// called with an invalid {} struct type or that the code generation is out'
            .format(current_api_data.api_name),
            file=self.outFile
        )
        write('// of date, both of which are errors', file=self.outFile)
        write(
            'template <typename T> {} Get{}() = delete;'.format(
                current_api_data.struct_type_enum,
                current_api_data.struct_type_func_prefix
            ),
            file=self.outFile
        )
        self.newline()

    def skip_struct_type(self, struct):
        """ Maybe be overridden """
        return struct in self.children_structs

    def write_struct_type_data(self):
        current_api_data = self.get_api_data()
        for struct in self.all_struct_members.keys():
            if self.skip_struct_type(struct):
                continue

            if struct in self.struct_type_names:
                write(
                    'template <> constexpr {} Get{}<{}>(){{ return {}; }}'.
                    format(
                        current_api_data.struct_type_enum,
                        current_api_data.struct_type_func_prefix, struct,
                        self.struct_type_names[struct]
                    ),
                    file=self.outFile
                )

        self.newline()



#!/usr/bin/env python
#
# Copyright 2007 Neal Norwitz
# Portions Copyright 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""C++ keywords and helper utilities for determining keywords."""

try:
    # Python 3.x
    import builtins
except ImportError:
    # Python 2.x
    import __builtin__ as builtins


if not hasattr(builtins, 'set'):
    # Nominal support for Python 2.3.
    from sets import Set as set


TYPES = set('bool char int long short double float void wchar_t unsigned signed'.split())
TYPE_MODIFIERS = set('auto register const inline extern static virtual volatile mutable'.split())
ACCESS = set('public protected private friend'.split())

CASTS = set('static_cast const_cast dynamic_cast reinterpret_cast'.split())

OTHERS = set('true false asm class namespace using explicit this operator sizeof'.split())
OTHER_TYPES = set('new delete typedef struct union enum typeid typename template'.split())

CONTROL = set('case switch default if else return goto'.split())
EXCEPTION = set('try catch throw'.split())
LOOP = set('while do for break continue'.split())

ALL = TYPES | TYPE_MODIFIERS | ACCESS | CASTS | OTHERS | OTHER_TYPES | CONTROL | EXCEPTION | LOOP


def IsKeyword(token):
    return token in ALL

def IsBuiltinType(token):
    if token in ('virtual', 'inline'):
        # These only apply to methods, they can't be types by themselves.
        return False
    return token in TYPES or token in TYPE_MODIFIERS



#!/usr/bin/env python3
# Copyright (c) 2022 Advanced Micro Devices, Inc. All rights reserved
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

# Helper script to perform automatic renaming of gfxrecon-optimize.exe prior to optimization


import sys
import os
import subprocess
import shutil
from os.path import exists


# Make sure we didn't leave stuff behind
def cleanup(optimizer_tool_path, optimizer_tool_path_renamed):
    if exists(optimizer_tool_path) and exists(optimizer_tool_path_renamed):
        optimizer_tool_file_size = os.path.getsize(optimizer_tool_path)
        optimizer_tool_renamed_file_size = os.path.getsize(optimizer_tool_path_renamed)
        if optimizer_tool_file_size == optimizer_tool_renamed_file_size:
            if "gfxrecon-replay.exe" not in optimizer_tool_path_renamed:
                print("Deleting: " + optimizer_tool_path_renamed)
                os.remove(optimizer_tool_path_renamed)


# Rename the optimizer
def rename_optimizer(encoded_app_executable):
    optimizer_name = os.path.join(os.path.dirname(__file__), "gfxrecon-optimize.exe")
    optimizer_name_renamed = os.path.join(os.getcwd(), encoded_app_executable)
    if exists(optimizer_name_renamed):
        print("Warning: " + encoded_app_executable +  " already exists. Reusing it.")
    else:
        shutil.copy(optimizer_name, optimizer_name_renamed)
        if not exists(capture_path):
            optimizer_name_renamed = ""
    return optimizer_name_renamed


# Run the optimizer
def run_optimizer(replay_tool_path, args):
    cmd = replay_tool_path
    cmd += " \"" + args.pop(0) + "\""
    cmd += " \"" + args.pop(0) + "\""

    for arg in args:
        cmd += " " + arg
    print("Running: " + cmd)
    print("")
    output = subprocess.check_output(cmd).decode().split("\r\n")
    for line in output:
        print(line)


# Run gfxrecon-info to retrieve application name from capture
def retrieve_exe_name(info_tool_path, capture_path):
    print("Retrieving captured application executable name...")
    cmd = info_tool_path + " \"" + capture_path + "\" --exe-info-only"
    output = subprocess.check_output(cmd).decode('utf-8',errors='ignore').split("\r\n")
    exe_name = ""
    for line in output:
        if "Application exe name:" in line:
            tokens = line.split(" ")
            for token in tokens:
                exe_name = token.strip()
    if ".exe" not in exe_name:
        exe_name = ""
    else:
        print("Found: " + exe_name)
    return exe_name


# Print usage instructions
def usage():
    print("gfxrecon-optimize-renamed.py - Helper script to perform automatic renaming of gfxrecon-optimize.exe prior to optimization.")
    print()
    print("Usage:")
    print("  gfxrecon-optimize-renamed.py <input-file> <output-file> [--dxr] [--d3d12-pso-removal]")
    print()
    print("Required arguments:")
    print("  <input-file>          The path to input GFXReconstruct capture file to be processed.")
    print("  <output-file>         The path to output GFXReconstruct capture file to be created.")    
    print()
    print("Optional arguments:")    
    print("  --d3d12-pso-removal   D3D12-only: Remove creation of unreferenced PSOs.")
    print("  --dxr                 D3D12-only: Optimize for DXR replay.")
    print("  --gpu                 D3D12-only: Use the specified device for the optimizer replay, where index is the zero-based index")
    print("                        to the array of physical devices returned by vkEnumeratePhysicalDevices or The optimizer replay")
    print("                        may fail if the specified device is not compatible with theIDXGIFactory1::EnumAdapters1.")
    print("                        The optimizer replay may fail if the specified device is not compatible with the original")
    print("                        capture devices.")
    print()    
    print("Note: running without optional arguments will instruct the optimizer to detect API and run all available optimizations.")
    print()
    print("Example manual usage (D3D12):")
    print("  gfxrecon-optimize-renamed.py my_capture.gfxr my_capture_dxr_optimized.gfxr --dxr")
    print("  gfxrecon-optimize-renamed.py my_capture.gfxr my_capture_pso_optimized.gfxr --d3d12-pso-removal")
    print()
    print("Example automatic usage (D3D12 + Vulkan):")
    print("  gfxrecon-optimize-renamed.py my_capture.gfxr my_capture_optimized.gfxr") 
    print();

# Main
if __name__ == '__main__':

    argc = len(sys.argv)

    # If valid number of params
    if argc > 2:
        # exe gfxr gfxr or exe gfxr gfxr optional
        if sys.argv[1][-4:] == "gfxr" and sys.argv[2][-4:] == "gfxr":
            capture_path = sys.argv[1]
        
            args = []
            for arg in sys.argv:
                args.append(arg)
            args.pop(0)

            optimizer_tool_path_renamed = ""

            try:
                # Verify capture exists
                if exists(capture_path):
                    optimizer_tool_path = os.path.join(os.path.dirname(__file__), "gfxrecon-optimize.exe")

                    # Verify optimizer tool exists
                    if exists(optimizer_tool_path):
                        info_tool_path = os.path.join(os.path.dirname(__file__), "gfxrecon-info.exe")

                        # Verify info tool exists
                        if exists(info_tool_path):
                            encoded_app_executable = retrieve_exe_name(info_tool_path, capture_path)

                            if encoded_app_executable:
                                optimizer_tool_path_renamed = rename_optimizer(encoded_app_executable)
                                if optimizer_tool_path_renamed:
                                    run_optimizer(optimizer_tool_path_renamed, args)
                                    cleanup(optimizer_tool_path, optimizer_tool_path_renamed)
                                else:
                                    print("Warning: Could not rename optimizer")
                                    run_optimizer(optimizer_tool_path, args)
                            else:
                                print("Warning: Did not detect captured application executable name")
                                run_optimizer(optimizer_tool_path, args)
                        else:
                            print("Error: ensure gfxrecon-info.exe lives in the same directory as this script")
                    else:
                        print("Error: ensure gfxrecon-optimize.exe lives in the same directory as this script")
                else:
                    usage()
                    print("Error: path to capture is invalid")
            except Exception as e:
                print("Error: exception occurred")
                print(e)
                cleanup(optimizer_tool_path, optimizer_tool_path_renamed)
        else:
            print("Error: path to capture is invalid")
            print("Usage: gfxrecon-optimize-renamed.py input_file.gfxr output_file.gfxr [--dxr] [--d3d12-pso-removal]")

    else:
        print("Error: missing path to capture")
        usage()
        



#!/usr/bin/env python3
# Copyright (c) 2017-2024 Google LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and/or associated documentation files (the
# "Materials"), to deal in the Materials without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Materials, and to
# permit persons to whom the Materials are furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Materials.
#
# MODIFICATIONS TO THIS FILE MAY MEAN IT NO LONGER ACCURATELY REFLECTS
# KHRONOS STANDARDS. THE UNMODIFIED, NORMATIVE VERSIONS OF KHRONOS
# SPECIFICATIONS AND HEADER INFORMATION ARE LOCATED AT
#    https://www.khronos.org/registry/
#
# THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.

"""Generates a C language headers from a SPIR-V JSON grammar file"""

import errno
import json
import os.path
import re

DEFAULT_COPYRIGHT="""Copyright (c) 2020-2024 The Khronos Group Inc.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and/or associated documentation files (the
"Materials"), to deal in the Materials without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Materials, and to
permit persons to whom the Materials are furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Materials.

MODIFICATIONS TO THIS FILE MAY MEAN IT NO LONGER ACCURATELY REFLECTS
KHRONOS STANDARDS. THE UNMODIFIED, NORMATIVE VERSIONS OF KHRONOS
SPECIFICATIONS AND HEADER INFORMATION ARE LOCATED AT
   https://www.khronos.org/registry/

THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.
""".split('\n')

def make_path_to_file(f):
    """Makes all ancestor directories to the given file, if they
    don't yet exist.

    Arguments:
        f: The file whose ancestor directories are to be created.
    """
    dir = os.path.dirname(os.path.abspath(f))
    try:
        os.makedirs(dir)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(dir):
            pass
        else:
            raise

class ExtInstGrammar:
    """The grammar for an extended instruction set"""

    def __init__(self, name, copyright, instructions, operand_kinds, version = None, revision = None):
       self.name = name
       self.copyright = copyright
       self.instructions = instructions
       self.operand_kinds = operand_kinds
       self.version = version
       self.revision = revision


class LangGenerator:
    """A language-specific generator"""

    def __init__(self):
        self.upper_case_initial = re.compile('^[A-Z]')
        pass

    def comment_prefix(self):
        return ""

    def namespace_prefix(self):
        return ""

    def uses_guards(self):
        return False

    def cpp_guard_preamble(self):
        return ""

    def cpp_guard_postamble(self):
        return ""

    def enum_value(self, prefix, name, value):
        if self.upper_case_initial.match(name):
            use_name = name
        else:
            use_name = '_' + name

        return "    {}{} = {},".format(prefix, use_name, value)

    def generate(self, grammar):
        """Returns a string that is the language-specific header for the given grammar"""

        parts = []
        if grammar.copyright:
            parts.extend(["{}{}".format(self.comment_prefix(), f) for f in grammar.copyright])
        parts.append('')

        guard = 'SPIRV_UNIFIED1_{}_H_'.format(grammar.name)
        if self.uses_guards:
            parts.append('#ifndef {}'.format(guard))
            parts.append('#define {}'.format(guard))
        parts.append('')

        parts.append(self.cpp_guard_preamble())

        if grammar.version:
            parts.append(self.const_definition(grammar.name, 'Version', grammar.version))

        if grammar.revision is not None:
            parts.append(self.const_definition(grammar.name, 'Revision', grammar.revision))

        parts.append('')

        if grammar.instructions:
            parts.append(self.enum_prefix(grammar.name, 'Instructions'))
            for inst in grammar.instructions:
                parts.append(self.enum_value(grammar.name, inst['opname'], inst['opcode']))
            parts.append(self.enum_end(grammar.name, 'Instructions'))
            parts.append('')

        if grammar.operand_kinds:
            for kind in grammar.operand_kinds:
                parts.append(self.enum_prefix(grammar.name, kind['kind']))
                for e in kind['enumerants']:
                    parts.append(self.enum_value(grammar.name, e['enumerant'], e['value']))
                parts.append(self.enum_end(grammar.name, kind['kind']))
            parts.append('')

        parts.append(self.cpp_guard_postamble())

        if self.uses_guards:
            parts.append('#endif // {}'.format(guard))

        # Ensre the file ends in an end of line
        parts.append('')

        return '\n'.join(parts)


class CLikeGenerator(LangGenerator):
    def uses_guards(self):
        return True

    def comment_prefix(self):
        return "// "

    def const_definition(self, prefix, var, value):
        # Use an anonymous enum.  Don't use a static const int variable because
        # that can bloat binary size.
        return 'enum {0}{1}{2}{3} = {4},{1}{2}{3}_BitWidthPadding = 0x7fffffff{5};'.format(
               '{', '\n    ', prefix, var, value, '\n}')

    def enum_prefix(self, prefix, name):
        return 'enum {}{} {}'.format(prefix, name, '{')

    def enum_end(self, prefix, enum):
        return '    {}{}Max = 0x7fffffff\n{};\n'.format(prefix, enum, '}')

    def cpp_guard_preamble(self):
        return '#ifdef __cplusplus\nextern "C" {\n#endif\n'

    def cpp_guard_postamble(self):
        return '#ifdef __cplusplus\n}\n#endif\n'


class CGenerator(CLikeGenerator):
    pass


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Generate language headers from a JSON grammar')

    parser.add_argument('--extinst-name',
                        type=str, required=True,
                        help='The name to use in tokens')
    parser.add_argument('--extinst-grammar', metavar='<path>',
                        type=str, required=True,
                        help='input JSON grammar file for extended instruction set')
    parser.add_argument('--extinst-output-base', metavar='<path>',
                        type=str, required=True,
                        help='Basename of the language-specific output file.')
    args = parser.parse_args()

    with open(args.extinst_grammar) as json_file:
        grammar_json = json.loads(json_file.read())
        if 'copyright' in grammar_json:
          copyright = grammar_json['copyright']
        else:
          copyright = DEFAULT_COPYRIGHT
        if 'version' in grammar_json:
          version = grammar_json['version']
        else:
          version = 0
        if 'operand_kinds' in grammar_json:
          operand_kinds = grammar_json['operand_kinds']
        else:
          operand_kinds = []

        grammar = ExtInstGrammar(name = args.extinst_name,
                                 copyright = copyright,
                                 instructions = grammar_json['instructions'],
                                 operand_kinds = operand_kinds,
                                 version = version,
                                 revision = grammar_json['revision'])
        make_path_to_file(args.extinst_output_base)
        with open(args.extinst_output_base + '.h', 'w') as f:
            f.write(CGenerator().generate(grammar))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2023 Valve Corporation
# Copyright (c) 2021, 2023 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from dx12_base_generator import Dx12BaseGenerator, write
from dx12_enum_to_string_header_generator import Dx12EnumToStringHeaderGenerator # For the list of substrings in bitflag

class Dx12EnumToStringBodyGenerator(Dx12BaseGenerator):
    """TODO : Generates C++ functions responsible for Convert to texts."""

    BITS_LIST = Dx12EnumToStringHeaderGenerator.BITS_LIST

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )

    def beginFile(self, gen_opts):
        """Method override."""
        Dx12BaseGenerator.beginFile(self, gen_opts)

        code = '#if defined(D3D12_SUPPORT)\n'
        code += '\n'
        code += '#include "generated_dx12_enum_to_string.h"\n'
        write(code, file=self.outFile)

        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(util)', file=self.outFile)
        self.newline()

    def generate_feature(self):
        for k, v in self.source_dict['enum_dict'].items():
            # Generate enum handler for all enums
            body = 'std::string ToString(const {0} value)\n'
            body += '{{\n'
            body += '    const char* ret = "Unhandled {0}";\n'
            body += '    switch (value) {{\n'
            processed_values = set()
            for value in v['values']:
                if not value['value'] in processed_values:
                    body += '        case {0}: ret = "{0}"; break;\n'.format(value['name'])
                    processed_values.add(value['name'])
                    processed_values.add(value['value'])
            body += '    }}\n'
            body += '    return ret;\n'
            body += '}}\n'

            # Generate flags handler for enums identified as bitmasks
            for bits in self.BITS_LIST:
                if k.find(bits) >= 0:
                    body += '\nstd::string ToString_{0}(const uint32_t flags)\n'
                    body += '{{\n'
                    body += '    return BitmaskToString<{0}>(flags);\n'
                    body += '}}\n'
            write(body.format(k), file=self.outFile)

        # Generate REFIID handler
        iids = list()
        for k, v in self.source_dict['header_dict'].items():
            if hasattr(v, 'variables'):
                for m in v.variables:
                    if 'DEFINE_GUID' in m['type']:
                        index = m['type'].find(',')
                        iids.append(m['type'][len('DEFINE_GUID ( '):index])
        body = 'std::string ToString(const IID& iid)\n'
        body += '{\n'
        if not "IID_IUnknown" in iids:
            iids.append("IID_IUnknown")
        for iid in iids:
            body += '    if (iid == {0}) return "{0}";\n'.format(iid)
        body += '    return "Invalid IID";\n'
        body += '}\n'
        write(body, file=self.outFile)

    def endFile(self):
        """Method override."""
        self.newline()
        body = 'GFXRECON_END_NAMESPACE(util)\n'
        body += 'GFXRECON_END_NAMESPACE(gfxrecon)\n'
        body += '\n'
        body += '#endif // defined(D3D12_SUPPORT)'
        write(body, file=self.outFile)

        # Finish processing in superclass
        Dx12BaseGenerator.endFile(self)



#!/usr/bin/python3 -i
#
# Copyright (c) 2019-2020 Valve Corporation
# Copyright (c) 2019-2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from khronos_base_generator import write


class KhronosStructHandleMappersHeaderGenerator():
    """Base class for generating struct handle mappers header code."""

    # Recursively search a structs members to see if they too belong in the
    # output struct list.  This could be because an including struct is an
    # output struct.
    def process_struct_members_to_output_struct(self, value):
        for member in self.all_struct_members[value.base_type]:
            if (
                self.is_struct(member.base_type)
                and not self.is_struct_black_listed(member.base_type)
                and (member.base_type not in self.all_struct_aliases)
                and (member.base_type in self.structs_with_handles)
                and (member.base_type not in self.output_structs)
            ):
                self.output_structs.append(member.base_type)
                self.process_struct_members_to_output_struct(member)

    def write_struct_handle_mapper_header(self):

        extended_struct_func_name = self.get_extended_struct_func_prefix()

        for struct in self.get_all_filtered_struct_names():
            if (
                (struct in self.structs_with_handles) or
                self.child_struct_has_handles(struct) or
                (struct in self.GENERIC_HANDLE_STRUCTS)
            ) and (struct not in self.STRUCT_MAPPERS_BLACKLIST):
                body = '\n'
                body += 'void MapStructHandles(Decoded_{}* wrapper, const CommonObjectInfoTable& object_info_table);'.format(
                    struct
                )
                write(body, file=self.outFile)

        self.newline()
        write(
            'void Map{func}StructHandles({node}Node* value, const CommonObjectInfoTable& object_info_table);'.format(
                    func=self.get_extended_struct_func_prefix(),
                    node=self.get_extended_struct_node_prefix()
                ),
            file=self.outFile
        )
        self.newline()

        # List of structs containing handles that are also used as output parameters for a command
        self.output_structs = []

        # Look for output structs that contain handles and add to list
        for cmd in self.get_all_filtered_cmd_names():
            for value_info in self.all_cmd_params[cmd][2]:
                if (
                    self.is_output_parameter(value_info)
                    and self.is_struct(value_info.base_type)
                    and not self.is_struct_black_listed(value_info.base_type)
                    and (value_info.base_type not in self.all_struct_aliases)
                    and (value_info.base_type in self.structs_with_handles)
                    and (value_info.base_type not in self.output_structs)
                ):
                    self.output_structs.append(value_info.base_type)
                    self.process_struct_members_to_output_struct(value_info)

        for struct in self.output_structs:
            write(
                'void AddStructHandles(format::HandleId parent_id, const Decoded_{type}* id_wrapper, const {type}* handle_struct, CommonObjectInfoTable* object_info_table);'
                .format(type=struct),
                file=self.outFile
            )
            self.newline()

        for struct in self.output_structs:
            if struct in self.structs_with_handle_ptrs:
                write(
                    'void SetStructHandleLengths(Decoded_{type}* wrapper);'.
                    format(type=struct),
                    file=self.outFile
                )
                self.newline()

        write(
            '#include "decode/common_struct_handle_mappers.h"',
            file=self.outFile
        )



# Copyright 2018-2024, The Khronos Group Inc.
# Copyright (c) 2018-2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Author(s):    Rylie Pavlik <rylie.pavlik@collabora.com>
"""Provides the BasePrinter base class for MacroChecker/Message output techniques."""

from abc import ABC, abstractmethod
from pathlib import Path

from .macro_checker import MacroChecker
from .macro_checker_file import MacroCheckerFile
from .shared import EntityData, Message, MessageContext, MessageType


def getColumn(message_context):
    """Return the (zero-based) column number of the message context.

    If a group is specified: returns the column of the start of the group.
    If no group, but a match is specified: returns the column of the start of
    the match.
    If no match: returns column 0 (whole line).
    """
    if not message_context.match:
        # whole line
        return 0
    if message_context.group is not None:
        return message_context.match.start(message_context.group)
    return message_context.match.start()


class BasePrinter(ABC):
    """Base class for a way of outputting results of a checker execution."""

    def __init__(self):
        """Constructor."""
        self._cwd = None
        self.show_script_location = False

    def close(self):
        """Write the tail end of the output and close it, if applicable.

        Override if you want to print a summary or are writing to a file.
        """
        pass

    ###
    # Output methods: these should all print/output directly.
    def output(self, obj):
        """Output any object.

        Delegates to other output* methods, if type known,
        otherwise uses self.outputFallback().
        """
        if isinstance(obj, Message):
            self.outputMessage(obj)
        elif isinstance(obj, MacroCheckerFile):
            self.outputCheckerFile(obj)
        elif isinstance(obj, MacroChecker):
            self.outputChecker(obj)
        else:
            self.outputFallback(self.formatBrief(obj))

    @abstractmethod
    def outputResults(self, checker, broken_links=True,
                      missing_includes=False):
        """Output the full results of a checker run.

        Must be implemented.

        Typically will call self.output() on the MacroChecker,
        as well as calling self.outputBrokenAndMissing()
        """
        raise NotImplementedError

    @abstractmethod
    def outputBrokenLinks(self, checker, broken):
        """Output the collection of broken links.

        `broken` is a dictionary of entity names: usage contexts.

        Must be implemented.

        Called by self.outputBrokenAndMissing() if requested.
        """
        raise NotImplementedError

    @abstractmethod
    def outputMissingIncludes(self, checker, missing):
        """Output a table of missing includes.

        `missing` is a iterable entity names.

        Must be implemented.

        Called by self.outputBrokenAndMissing() if requested.
        """
        raise NotImplementedError

    def outputChecker(self, checker):
        """Output the contents of a MacroChecker object.

        Default implementation calls self.output() on every MacroCheckerFile.
        """
        for f in checker.files:
            self.output(f)

    def outputCheckerFile(self, fileChecker):
        """Output the contents of a MacroCheckerFile object.

        Default implementation calls self.output() on every Message.
        """
        for m in fileChecker.messages:
            self.output(m)

    def outputBrokenAndMissing(self, checker, broken_links=True,
                               missing_includes=False):
        """Outputs broken links and missing includes, if desired.

        Delegates to self.outputBrokenLinks() (if broken_links==True)
        and self.outputMissingIncludes() (if missing_includes==True).
        """
        if broken_links:
            broken = checker.getBrokenLinks()
            if broken:
                self.outputBrokenLinks(checker, broken)
        if missing_includes:
            missing = checker.getMissingUnreferencedApiIncludes()
            if missing:
                self.outputMissingIncludes(checker, missing)

    @abstractmethod
    def outputMessage(self, msg):
        """Output a Message.

        Must be implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def outputFallback(self, msg):
        """Output some text in a general way.

        Must be implemented.
        """
        raise NotImplementedError

    ###
    # Format methods: these should all return a string.
    def formatContext(self, context, _message_type=None):
        """Format a message context in a verbose way, if applicable.

        May override, default implementation delegates to
        self.formatContextBrief().
        """
        return self.formatContextBrief(context)

    def formatContextBrief(self, context, _with_color=True):
        """Format a message context in a brief way.

        May override, default is relativeFilename:line:column
        """
        relative = self.getRelativeFilename(context.filename)
        return f'{relative}:{context.lineNum}:{getColumn(context)}'

    def formatMessageTypeBrief(self, message_type, _with_color=True):
        """Format a message type in a brief way.

        May override, default is message_type:
        """
        return f'{message_type}:'

    def formatEntityBrief(self, entity_data, _with_color=True):
        """Format an entity in a brief way.

        May override, default is macro:entity.
        """
        return f'{entity_data.macro}:{entity_data.entity}'

    def formatBrief(self, obj, with_color=True):
        """Format any object in a brief way.

        Delegates to other format*Brief methods, if known,
        otherwise uses str().
        """
        if isinstance(obj, MessageContext):
            return self.formatContextBrief(obj, with_color)
        if isinstance(obj, MessageType):
            return self.formatMessageTypeBrief(obj, with_color)
        if isinstance(obj, EntityData):
            return self.formatEntityBrief(obj, with_color)
        return str(obj)

    @property
    def cwd(self):
        """Get the current working directory, fully resolved.

        Lazy initialized.
        """
        if not self._cwd:
            self._cwd = Path('.').resolve()
        return self._cwd

    ###
    # Helper function
    def getRelativeFilename(self, fn):
        """Return the given filename relative to the current directory,
        if possible.
        """
        try:
            return str(Path(fn).relative_to(self.cwd))
        except ValueError:
            return str(Path(fn))



#!/usr/bin/python3 -i
#
# Copyright (c) 2018 Valve Corporation
# Copyright (c) 2018-2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from khronos_base_generator import write


class KhronosStructDecodersForwardGenerator():
    """KhronosStructDecodersForwardGenerator
    Generates C++ type and function declarations for decoding Khronos API structures.
    """

    def write_struct_decoder_forward_prototypes(self):
        for struct in self.get_all_filtered_struct_names():
            write('struct Decoded_{};'.format(struct), file=self.outFile)

        self.newline()

        for struct in self.get_all_filtered_struct_names():
            write(
                'size_t DecodeStruct(const uint8_t* parameter_buffer, size_t buffer_size, Decoded_{}* wrapper);'
                .format(struct),
                file=self.outFile
            )



#!/usr/bin/python3 -i
#
# Copyright (c) 2022-2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

from khronos_base_generator import write
from reformat_code import indent_cpp_code, remove_leading_empty_lines, remove_trailing_newlines

class KhronosStructToJsonBodyGenerator():
    """KhronosStructToJsonBodyGenerator
    Generate C++ function definitions to serialize Khronos structures to JSON"""

    def should_decode_struct(self, struct):
        """Method indended to be overridden.
        Indicates that the provided struct is a struct we want to decode"""
        return True

    def get_local_type_var_name(self):
        return 's_type'

    def decode_as_handle(self, parent_type, member):
        """ Method which may be overridden.
        Indicates that the given type should be decoded as a handle."""
        return self.is_handle(member.base_type)

    def write_body_contents(self):
        write('using util::JsonOptions;', file=self.outFile)
        write('using util::to_hex_variable_width;', file=self.outFile)
        write('using util::uuid_to_string;', file=self.outFile)
        self.newline()

        for struct in self.get_all_filtered_struct_names():
            if self.should_decode_struct(struct):
                body = indent_cpp_code('''
                    void FieldToJson(nlohmann::ordered_json& jdata, const Decoded_{0}* data, const JsonOptions& options)
                    {{
                        if (data && data->decoded_value)
                        {{
                            const {0}& decoded_value = *data->decoded_value;
                            const Decoded_{0}& meta_struct = *data;
                    '''.format(struct))

                if struct in self.children_structs:
                    body += self.make_base_struct_body(struct)

                body += self.makeStructBody(
                    struct, self.all_struct_members[struct]
                )

                body += '    }\n'
                body += '}\n'

                body = remove_trailing_newlines(body)
                #body = remove_trailing_newlines(indent_cpp_code(body))
                write(body, file=self.outFile)

        stype_var = self.get_struct_type_var_name()
        extended_node_prefix = self.get_extended_struct_node_prefix()
        stype_auto = self.get_local_type_var_name()
        base_in_struct = self.get_base_input_structure_name()
        
        body = '''
            void FieldToJson(nlohmann::ordered_json& jdata, const {}Node* data, const JsonOptions& options)
            {{
                if (data && data->GetPointer())
                {{
        '''.format(extended_node_prefix)
        body += '            const auto {} = reinterpret_cast<const {}*>(data->GetPointer())->{};'.format(
            stype_auto, base_in_struct, stype_var
        )
        body += '''
                    switch ({})
                    {{'''.format(stype_auto)
        body += self.make_extended_struct_body()
        body = indent_cpp_code(body)
        write(body, file=self.outFile)

    def make_base_struct_body(
            self,
            base_struct_type,
            indent = '    ',
            decoded_value='decoded_value',
            data='data',
            json_data='jdata',
            json_options='options'):
        # Otherwise, we need to go through and actually decode the appropriate
        # type of the struct pointed at by the base header struct pointer.
        body = ''
        indent1 = indent + '    '
        indent2 = indent1 + '    '

        body += f'{indent1}switch ({decoded_value}.type)\n'
        body += f'{indent1}{{\n'
        body += f'{indent2}default:\n'
        body += f'{indent2}    // Handle as base-type below\n'
        body += f'{indent2}    break;\n'

        for child in self.children_structs[base_struct_type]:
            struct_type = self.struct_type_names[child]
            body += f'{indent2}case {struct_type}:\n'
            body += f'{indent2}    FieldToJson({json_data},\n'
            body += f'{indent2}                reinterpret_cast<const Decoded_{child}*>({data}),\n'
            body += f'{indent2}                {json_options});\n'
            body += f'{indent2}    // Return here because we processed the appropriate data in\n'
            body += f'{indent2}    // the correct structure type\n'
            body += f'{indent2}    return;\n'

        body += f'{indent1}}}\n\n'

        return body


    #
    # Command definition
    def makeStructBody(self, name, values):
        body = ''
        has_extended_struct = False
        extended_struct_var_name = self.get_extended_struct_var_name()
        for value in values:
            type_name = self.make_decoded_param_type(value)

            # Determine the appropriate type.  If it's an alias, get
            # the original type
            value_type = value.base_type
            if (self.is_struct(value.base_type) and value.base_type in self.all_struct_aliases):
                value_type = self.all_struct_aliases[value.base_type]
            flagsEnumType = value_type

            if value.name == extended_struct_var_name:
                has_extended_struct = True
                continue

            # Default to getting the data from the native Vulkan struct:
            to_json = 'FieldToJson(jdata["{0}"], decoded_value.{0}, options)'

            if (
                self.is_function_ptr(value_type)
                or ('pUserData' == value.name or 'userData' == value.name)
            ):
                to_json = 'FieldToJson(jdata["{0}"], to_hex_variable_width(meta_struct.{0}), options)'
            elif value.is_pointer:
                if 'String' in type_name:
                    to_json = 'FieldToJson(jdata["{0}"], &meta_struct.{0}, options)'
                elif self.is_handle_like(value_type):
                    to_json = 'HandleToJson(jdata["{0}"], &meta_struct.{0}, options)'
                else:
                    to_json = 'FieldToJson(jdata["{0}"], meta_struct.{0}, options)'
            else:
                if value.is_array:
                    if 'UUID' in value.array_length or 'LUID' in value.array_length:
                        to_json = 'FieldToJson(jdata["{0}"], uuid_to_string(sizeof(decoded_value.{0}), decoded_value.{0}), options)'
                    elif 'String' in type_name:
                        to_json = 'FieldToJson(jdata["{0}"], &meta_struct.{0}, options)'
                    elif self.is_handle_like(value_type):
                        to_json = 'HandleToJson(jdata["{0}"], &meta_struct.{0}, options)'
                    elif self.is_struct(value_type):
                        # If this is a parent class, generate the parent->child conversion info
                        # appropriately
                        if value_type in self.children_structs.keys():
                            to_json = 'ParentChildFieldToJson(args["{0}"], {0}, json_options)'
                        else:
                            to_json = 'FieldToJson(jdata["{0}"], meta_struct.{0}, options)'
                    elif not value.is_dynamic:
                        to_json = 'FieldToJson(jdata["{0}"], &meta_struct.{0}, options)'
                    else:
                        to_json = 'FieldToJson(jdata["{0}"], meta_struct.{0}, options)'
                else:
                    if self.decode_as_handle(name, value):
                        to_json = 'HandleToJson(jdata["{0}"], meta_struct.{0}, options)'
                    elif value_type in self.formatAsHex:
                        to_json = 'FieldToJson(jdata["{0}"], to_hex_variable_width(decoded_value.{0}), options)'
                    elif self.is_struct(value_type):
                        to_json = 'FieldToJson(jdata["{0}"], meta_struct.{0}, options)'
                    elif self.is_flags(value_type):
                        if value_type in self.flags_type_aliases:
                            flagsEnumType = self.flags_type_aliases[
                                value_type]
                        to_json = 'FieldToJson({2}_t(),jdata["{0}"], decoded_value.{0}, options)'
                    elif self.is_enum(value_type):
                        to_json = 'FieldToJson(jdata["{0}"], decoded_value.{0}, options)'
                    elif self.is_boolean_type(value_type):
                        to_json = 'jdata["{0}"] = static_cast<bool>(decoded_value.{0})'

            to_json = to_json.format(
                value.name, value_type, flagsEnumType
            )
            body += '        {0};\n'.format(to_json)

        # Handle the extended struct last
        if has_extended_struct:
            body += '        FieldToJson(jdata["{0}"], meta_struct.{0}, options);\n'.format(
                extended_struct_var_name
            )

        return body

    def make_extended_struct_body(self):
        body = ''
        var_name = self.get_extended_struct_var_name().lower()
        stype_auto = self.get_local_type_var_name()

        extended_list = []
        for struct in self.all_extended_structs:
            for ext_struct in self.all_extended_structs[struct]:
                if ext_struct not in extended_list and ext_struct not in self.all_struct_aliases:
                    extended_list.append(ext_struct)

        for struct in sorted(extended_list):
            if struct not in self.struct_type_names:
                continue
            body += '''
            case {1}:
            {{
               const auto* {2} = reinterpret_cast<const Decoded_{0}*>(data->GetMetaStructPointer());
               FieldToJson(jdata, {2}, options);
               break;
            }}
            '''.format(struct, self.struct_type_names[struct], var_name)
        body += '\n'
        body += '            default:\n'
        body += '            {\n'
        body += '                GFXRECON_LOG_WARNING("Unknown {} node type: %u.", (unsigned) {});\n'.format(
            var_name, stype_auto
        )
        body += '            }\n'
        body += '        }\n'
        body += '    }\n'
        body += '}\n'
        return body
